# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re

import livelex.regex


default_action = object()
default_target = object()


class Lexicon:
    """A Lexicon consists of a set of pattern rules a text is scanned for.
    
    """
    __slots__ = ('rules_func', 'lexicons', 're_flags')
    
    def __init__(self, rules_func,
                       re_flags=0,
        ):
        """Initializes with the rules function.
        
        The rules function accepts the Language class as argument, and yields
        the pattern, action, target, ... tuples.
        
        """
        self.rules_func = rules_func
        self.re_flags = re_flags
        self.lexicons = {}
    
    def __get__(self, instance, owner):
        """Called when accessed as a descriptor, via the Language class."""
        if instance:
            raise RuntimeError('Language should never be instantiated')
        try:
            lexicon = self.lexicons[owner]
        except KeyError:
            lexicon = self.lexicons[owner] = BoundLexicon(self, owner)
        return lexicon


class BoundLexicon:
    """A Bound Lexicon is tied to a particular class.
    
    This makes it possible to inherit from a Language class and change
    only some Lexicons.
    
    """
    __slots__ = ('lexicon', 'language', '_parser_func')
    
    def __init__(self, lexicon, language):
        self.lexicon = lexicon
        self.language = language

    def __call__(self):
        """Call the original function, yielding the rules."""
        return self.lexicon.rules_func(self.language)
    
    def name(self):
        """Return the 'Language.lexicon' name of this bound lexicon."""
        return '.'.join((self.language.__name__, self.lexicon.rules_func.__name__))

    @property
    def parse(self):
        try:
            f = self._parser_func
        except AttributeError:
            f = self._parser_func = self._get_parser_func()
        return f
    
    @property
    def re_flags(self):
        """The re_flags set on instantiation."""
        return self.lexicon.re_flags

    def _get_parser_func(self):
        """Compile the pattern rules and return a parse(text, pos) func."""
        patterns = []
        action_targets = []
        _default_action = None
        _default_target = None
        # make lists of pattern, action and possible targets
        for pattern, action, *target in self():
            if pattern is default_action:
                _default_action = action
            elif pattern is default_target:
                _default_target = action, *target
            else:
                if isinstance(pattern, livelex.regex.RegexBuilder):
                    pattern = pattern.build()
                patterns.append(pattern)
                action_targets.append((action, *target))
        # compile the regexp for all patterns
        rx = re.compile("|".join("(?P<g_{0}>{1})".format(i, pattern)
            for i, pattern in enumerate(patterns)), self.re_flags)
        # make a fast mapping list from matchObj.lastindex to the targets
        indices = sorted(v for k, v in rx.groupindex.items() if k.startswith('g_'))
        index = [None] * (indices[-1] + 1)
        for i, action_target in zip(indices, action_targets):
            index[i] = action_target
        
        if _default_action:
            if _default_target:
                raise RuntimeError(
                    "Lexicon {0}: can't have both default_action and "
                    "default_target".format(self.name()))
            def parse(text, pos):
                """Parse text, using a default action for unknown text."""
                for m in rx.finditer(text, pos):
                    if m.start() > pos:
                        yield pos, text[pos:m.start()], None, _default_action
                    yield (m.start(), m.group(), m, *index[m.lastindex])
                    pos = m.end()
                if pos < len(text):
                    yield (pos, text[pos:], None, _default_action)
        elif _default_target:
            def parse(text, pos):
                """Parse text, stopping with the default target at unknown text."""
                while True:
                    m = rx.match(text, pos)
                    if m:
                        yield (pos, m.group(), m, *index[m.lastindex])
                        pos = m.end()
                    else:
                        yield (pos, "", None, None, *_default_target)
                        break
        else:
            def parse(text, pos):
                """Parse text, skipping unknown text."""
                for m in rx.finditer(text, pos):
                    yield (m.start(), m.group(), m, *index[m.lastindex])
        return parse


def lexicon(rules_func=None, **kwargs):
    """Decorator to make a function in a Language class definition a Lexicon object."""
    if rules_func and not kwargs:
        return Lexicon(rules_func)
    def lexicon(rules_func):
        return Lexicon(rules_func, **kwargs)
    return lexicon






# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The action module defines Action and its subclasses.

If an instance of Action is encountered in a rule, it is called to return the 
desired action. Nesting is possible in most cases, only some actions require 
the match object to be present; and such actions can't be used as default
action, or inside subgroup_actions.

"""


class Action:
    def __init__(self, *args):
        self.args = args
    
    def filter_actions(self, lexer, pos, text, match):
        raise NotImplementedError


class SkipAction(Action):
    def __init__(self):
        pass

    def filter_actions(self, lexer, pos, text, match):
        yield from ()


class Subgroup(Action):
    """Yield actions from subgroups in a match."""
    def filter_actions(self, lexer, pos, text, match):
        for i, action in enumerate(self.args, match.lastindex + 1):
            yield from lexer.filter_actions(action, match.start(i), match.group(i), None)
        

class Match(Action):
    """Expects a function as argument that is called with the match object.
    
    The function should return the desired action.
    
    """
    def __init__(self, func):
        self.func = func

    def filter_actions(self, lexer, pos, text, match):
        action = self.func(match)
        yield from lexer.filter_actions(action, pos, text, match)


class Text(Action):
    """Expects a function as argument that is called with the matched text.
    
    The function should return the desired action.
    
    """
    def __init__(self, func):
        self.func = func

    def filter_actions(self, lexer, pos, text, match):
        action = self.func(text)
        yield from lexer.filter_actions(action, pos, text, None)


# used to suppress a token
skip = SkipAction()



# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The livelex Python module.

"""

import re


from .action import (
    Action,
    Subgroup,
    Match,
    Text,
    skip,
)

from .lexicon import (
    Lexicon,
    lexicon,
    default_action,
    default_target,
)

from .language import (
    Language,
)

from .lex import (
    Lexer,
)

from .regex import (
    Pattern,
    Words,
)


def version():
    from . import pkginfo
    return tuple(map(int, re.findall(r'\d+', pkginfo.version)))





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from livelex import *


class Language:
    """A Language represents a set of Lexicons comprising a specific language.
    
    A Language is never instantiated. The class itself serves as a namespace
    and can be inherited from.
    
    
    
    """

    @lexicon(re_flags=0)
    def root(cls):
        yield Words(('bla', 'BLA')), 'bla action'
        yield r'ble', 'ble action'
        yield r'\s+', skip      # this text is skipped
        yield r'(bl)(ub)', Subgroup('bl act', 'ub act')
        yield r'blo', 'blo action', cls.blo
        yield default_action, "TEXT"
    
    @lexicon
    def blo(cls):
        yield r'\s+', skip      # this text is skipped
        yield r'1', '1 in blo'
        yield r'4', '4 in blo, end', -1
        yield r'[0-9]', Text(lambda t: "has 3" if '3' in t else 'no 3')
        yield default_action, "unparsed"


        



# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import itertools


from livelex.action import Action, skip


class Lexer:
        
    def __init__(self, root_lexicon):
        self.root_lexicon = root_lexicon
    
    def lex(self, text, state=None, pos=0):
        """Yield tokens for the text.
        
        state is a list of context instances or names(?), the last one is the current
        context.
        
        """
    
        if state is None:
            state = []
        lexicon = self.get_lexicon(state)
        state_change = False
        while True:
            for pos, txt, match, action, *target in lexicon.parse(text, pos):
                if target:
                    state_change = True
                    lexicon = self.get_lexicon(state, target)
                if txt:
                    tokens = list(self.filter_actions(action, pos, txt, match))
                    if tokens:
                        for token in tokens[:-1]:
                            yield (*token, None)
                        for token in tokens[-1:]:
                            yield (*token, state_change)
                        state_change = False
                    pos += len(txt)
                if target:
                    break # continue with new lexicon
            else:
                break
        
    def get_lexicon(self, state, target=()):
        """Modify the state according to target and return the topmost lexicon."""
        for t in target:
            if type(t) is int:
                if t < 0:
                    del state[t:]
                elif t > 0:
                    state.extend(itertools.repeat(state[-1], t))
            else:
                state.append(t)
        if not state:
            state.append(self.root_lexicon)
        return state[-1]

    def filter_actions(self, action, pos, txt, match):
        """Handle filtering via Action instances."""
        if isinstance(action, Action):
            yield from action.filter_actions(self, pos, txt, match)
        else:
            yield pos, txt, action


